@@ -1,5 +1,3 @@
-/* $XdotOrg: app/xdm/greeter/Login.c,v 1.6 2006/06/03 00:05:24 alanc Exp $ */
-/* $Xorg: Login.c,v 1.4 2001/02/09 02:05:41 xorgcvs Exp $ */
 /*
 
 Copyright 1988, 1998  The Open Group
@@ -27,35 +25,29 @@ other dealings in this Software without prior written authorization
 from The Open Group.
 
 */
-/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved.
+/*
+ * Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the
- * "Software"), to deal in the Software without restriction, including
- * without limitation the rights to use, copy, modify, merge, publish,
- * distribute, and/or sell copies of the Software, and to permit persons
- * to whom the Software is furnished to do so, provided that the above
- * copyright notice(s) and this permission notice appear in all copies of
- * the Software and that both the above copyright notice(s) and this
- * permission notice appear in supporting documentation.
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
- * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
- * OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
- * HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
- * INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
- * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
- * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
  *
- * Except as contained in this notice, the name of a copyright holder
- * shall not be used in advertising or otherwise to promote the sale, use
- * or other dealings in this Software without prior written authorization
- * of the copyright holder.
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
-/* $XFree86: xc/programs/xdm/greeter/Login.c,v 3.17tsi Exp $ */
 
 /*
  * xdm - display manager daemon
@@ -64,33 +56,38 @@ from The Open Group.
  * Login.c
  */
 
-# include <X11/IntrinsicP.h>
-# include <X11/StringDefs.h>
-# include <X11/keysym.h>
-# include <X11/DECkeysym.h>
-# include <X11/Xfuncs.h>
+#include <X11/IntrinsicP.h>
+#include <X11/StringDefs.h>
+#include <X11/keysym.h>
+#include <X11/DECkeysym.h>
+#include <X11/Xfuncs.h>
 
-# include <stdio.h>
-# include <ctype.h>
-# include <time.h>
+#include <stdio.h>
+#include <ctype.h>
+#include <time.h>
+#include <string.h>
 
-# include "dm.h"
-# include "dm_error.h"
-# include "greet.h"
-# include "LoginP.h"
+#include "dm.h"
+#include "dm_error.h"
+#include "greet.h"
+#include "LoginP.h"
 
 #ifdef XPM
-#include <sys/stat.h>
-#include <unistd.h>
-#include <X11/Xlib.h>
-#include <X11/xpm.h>
-#include <X11/extensions/shape.h>
+# include <sys/stat.h>
+# include <unistd.h>
+# include <X11/Xlib.h>
+# include <X11/xpm.h>
+# include <X11/extensions/shape.h>
 #endif /* XPM */
 
 #include <X11/cursorfont.h>
 
 #ifdef USE_XINERAMA
-#include <X11/extensions/Xinerama.h>
+# include <X11/extensions/Xinerama.h>
+#endif
+
+#ifdef USE_XFT
+# include <X11/extensions/Xrender.h>
 #endif
 
 #ifndef DEBUG
@@ -125,7 +122,7 @@ static XtResource resources[] = {
 	goffset(x), XtRImmediate,	(XtPointer) -1},
     {XtNy, XtCY, XtRPosition, sizeof (Position),
 	goffset(y), XtRImmediate,	(XtPointer) -1},
-#ifndef USE_XFT    
+#ifndef USE_XFT
     {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
         offset(textpixel), XtRString,	XtDefaultForeground},
     {XtNpromptColor, XtCForeground, XtRPixel, sizeof(Pixel),
@@ -143,7 +140,7 @@ static XtResource resources[] = {
 /* end (caolan) */
 #endif /* DANCING */
 
-#ifdef XPM    
+#ifdef XPM
 /* added by Ivan Griffin (ivan.griffin@ul.ie) */
         {XtNlogoFileName, XtCLogoFileName, XtRString, sizeof(char*),
         offset(logoFileName), XtRImmediate, (XtPointer)0},
@@ -159,6 +156,8 @@ static XtResource resources[] = {
 	offset(hipixel), XtRString,	XtDefaultForeground},
     {XtNshdColor, XtCForeground, XtRPixel, sizeof (Pixel),
 	offset(shdpixel), XtRString,	XtDefaultForeground},
+    {XtNinpColor, XtCForeground, XtRPixel, sizeof (Pixel),
+	offset(inppixel), XtRString,	XtDefaultBackground},
     {XtNframeWidth, XtCFrameWidth, XtRInt, sizeof(int),
         offset(outframewidth), XtRImmediate, (XtPointer) 1},
     {XtNinnerFramesWidth, XtCFrameWidth, XtRInt, sizeof(int),
@@ -167,13 +166,13 @@ static XtResource resources[] = {
         offset(sepwidth), XtRImmediate, (XtPointer) 1},
 /* end (amit) */
 
-#ifndef USE_XFT    
+#ifndef USE_XFT
     {XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
-    	offset (textFont), XtRString,	"*-new century schoolbook-medium-r-normal-*-180-*"},
+	offset (textFont), XtRString,	"*-new century schoolbook-medium-r-normal-*-180-*"},
     {XtNpromptFont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
-    	offset (promptFont), XtRString, "*-new century schoolbook-bold-r-normal-*-180-*"},
+	offset (promptFont), XtRString, "*-new century schoolbook-bold-r-normal-*-180-*"},
     {XtNgreetFont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
-    	offset (greetFont), XtRString,	"*-new century schoolbook-bold-i-normal-*-240-*"},
+	offset (greetFont), XtRString,	"*-new century schoolbook-bold-i-normal-*-240-*"},
     {XtNfailFont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
 	offset (failFont), XtRString,	"*-new century schoolbook-bold-r-normal-*-180-*"},
 #else /* USE_XFT */
@@ -195,7 +194,7 @@ static XtResource resources[] = {
 	offset(failcolor), XtRString,	XtDefaultForeground},
 #endif
     {XtNgreeting, XtCGreeting, XtRString, sizeof (char *),
-    	offset(greeting), XtRString, "X Window System"},
+	offset(greeting), XtRString, "X Window System"},
     {XtNunsecureGreeting, XtCGreeting, XtRString, sizeof (char *),
 	offset(unsecure_greet), XtRString, "This is an unsecure session"},
     {XtNnamePrompt, XtCNamePrompt, XtRString, sizeof (char *),
@@ -204,11 +203,7 @@ static XtResource resources[] = {
 	offset(passwdPrompt), XtRString, "Password:  "},
     {XtNfail, XtCFail, XtRString, sizeof (char *),
 	offset(failMsg), XtRString,
-#if defined(sun) && defined(SVR4)
-     "Login incorrect or not on system console if root"
-#else
-     "Login incorrect"
-#endif
+	"Login incorrect or forbidden by policy"
     },
     {XtNchangePasswdMessage, XtCChangePasswdMessage, XtRString,
 	sizeof (char *), offset(passwdChangeMsg), XtRString,
@@ -226,7 +221,11 @@ static XtResource resources[] = {
     {XtNallowNullPasswd, XtCAllowNullPasswd, XtRBoolean, sizeof (Boolean),
 	offset(allow_null_passwd), XtRImmediate, (XtPointer) False},
     {XtNallowRootLogin, XtCAllowRootLogin, XtRBoolean, sizeof(Boolean),
-     offset(allow_root_login), XtRImmediate, (XtPointer) True}
+	offset(allow_root_login), XtRImmediate, (XtPointer) True},
+    {XtNechoPasswd, XtCEchoPasswd, XtRBoolean, sizeof(Boolean),
+	offset(echo_passwd), XtRImmediate, (XtPointer) False},
+    {XtNechoPasswdChar, XtCEchoPasswdChar, XtRString,	sizeof (char *),
+	offset(echo_passwd_char), XtRString, (XtPointer) "*" }
 };
 
 #undef offset
@@ -242,33 +241,33 @@ static XtResource resources[] = {
 # define F_DESCENT(f)	((w)->login.f##Font->max_bounds.descent)
 #endif
 
-# define TEXT_X_INC(w)		F_MAX_WIDTH(text)
-# define TEXT_Y_INC(w)		(F_ASCENT(text) + F_DESCENT(text))
+#define TEXT_X_INC(w)		F_MAX_WIDTH(text)
+#define TEXT_Y_INC(w)		(F_ASCENT(text) + F_DESCENT(text))
 
-# define PROMPT_X_INC(w)	F_MAX_WIDTH(prompt)
-# define PROMPT_Y_INC(w)	(F_ASCENT(prompt) + F_DESCENT(prompt))
+#define PROMPT_X_INC(w)	F_MAX_WIDTH(prompt)
+#define PROMPT_Y_INC(w)	(F_ASCENT(prompt) + F_DESCENT(prompt))
 
-# define GREET_X_INC(w)		F_MAX_WIDTH(greet)
-# define GREET_Y_INC(w)		(F_ASCENT(greet) + F_DESCENT(greet))
+#define GREET_X_INC(w)		F_MAX_WIDTH(greet)
+#define GREET_Y_INC(w)		(F_ASCENT(greet) + F_DESCENT(greet))
 
-# define FAIL_X_INC(w)		F_MAX_WIDTH(fail)
-# define FAIL_Y_INC(w)		(F_ASCENT(fail) + F_DESCENT(fail))
+#define FAIL_X_INC(w)		F_MAX_WIDTH(fail)
+#define FAIL_Y_INC(w)		(F_ASCENT(fail) + F_DESCENT(fail))
 
-# define Y_INC(w)	max (TEXT_Y_INC(w), PROMPT_Y_INC(w))
+#define Y_INC(w)	max (TEXT_Y_INC(w), PROMPT_Y_INC(w))
 
 
-# define PROMPT_TEXT(w,n) 	((w)->login.prompts[n].promptText)
-# define DEF_PROMPT_TEXT(w,n) 	((w)->login.prompts[n].defaultPrompt)
-# define VALUE_TEXT(w,n) 	((w)->login.prompts[n].valueText)
-# define VALUE_TEXT_MAX(w,n)	((w)->login.prompts[n].valueTextMax)
-# define VALUE_SHOW_START(w,n)	((w)->login.prompts[n].valueShownStart)
-# define VALUE_SHOW_END(w,n)	((w)->login.prompts[n].valueShownEnd)
-# define PROMPT_STATE(w,n) 	((w)->login.prompts[n].state)
-# define PROMPT_CURSOR(w,n)	((w)->login.prompts[n].cursor)
+#define PROMPT_TEXT(w,n) 	((w)->login.prompts[n].promptText)
+#define DEF_PROMPT_TEXT(w,n) 	((w)->login.prompts[n].defaultPrompt)
+#define VALUE_TEXT(w,n) 	((w)->login.prompts[n].valueText)
+#define VALUE_TEXT_MAX(w,n)	((w)->login.prompts[n].valueTextMax)
+#define VALUE_SHOW_START(w,n)	((w)->login.prompts[n].valueShownStart)
+#define VALUE_SHOW_END(w,n)	((w)->login.prompts[n].valueShownEnd)
+#define PROMPT_STATE(w,n) 	((w)->login.prompts[n].state)
+#define PROMPT_CURSOR(w,n)	((w)->login.prompts[n].cursor)
 
-# define CUR_PROMPT_CURSOR(w)	PROMPT_CURSOR(w,w->login.activePrompt)
+#define CUR_PROMPT_CURSOR(w)	PROMPT_CURSOR(w,w->login.activePrompt)
 
-# define CUR_PROMPT_TEXT(w, n)	(PROMPT_TEXT(w,n) != NULL ? \
+#define CUR_PROMPT_TEXT(w, n)	(PROMPT_TEXT(w,n) != NULL ? \
 				 PROMPT_TEXT(w,n) : DEF_PROMPT_TEXT(w,n))
 
 #ifdef USE_XFT
@@ -297,59 +296,58 @@ XmuXftTextWidth(Display *dpy, XftFont *font, FcChar8 *string, int len);
 #endif
 
 
-# define STRING_WIDTH(f, s) 	TEXT_WIDTH (f, s, strlen(s))
+#define STRING_WIDTH(f, s) 	TEXT_WIDTH (f, s, strlen(s))
 
+/* Padded width of logo image, if compiled with XPM support */
+#ifdef XPM
+# define LOGO_W(w)     ((w)->login.logoWidth + ((w)->login.logoPadding * 2))
+#else
+# define LOGO_W(w)     0
+#endif
 
+#define TEXT_PROMPT_W(w, m) (STRING_WIDTH(prompt, m) + w->login.inframeswidth)
 
-# define TEXT_PROMPT_W(w, m) (STRING_WIDTH(prompt, m) + w->login.inframeswidth)
-
-# define DEF_PROMPT_W(w,n) TEXT_PROMPT_W(w, w->login.prompts[n].defaultPrompt)
-# define CUR_PROMPT_W(w,n)  (max(MAX_DEF_PROMPT_W(w), PROMPT_TEXT(w,n) ? \
+#define DEF_PROMPT_W(w,n) TEXT_PROMPT_W(w, w->login.prompts[n].defaultPrompt)
+#define CUR_PROMPT_W(w,n)  (max(MAX_DEF_PROMPT_W(w), PROMPT_TEXT(w,n) ? \
 		     TEXT_PROMPT_W(w, PROMPT_TEXT(w,n)) : 0))
 
-# define MAX_DEF_PROMPT_W(w) (max(DEF_PROMPT_W(w,0), DEF_PROMPT_W(w,1)))
+#define MAX_DEF_PROMPT_W(w) (max(DEF_PROMPT_W(w,0), DEF_PROMPT_W(w,1)))
 
-# define GREETING(w)	((w)->login.secure_session  && !(w)->login.allow_access ?\
+#define GREETING(w)	((w)->login.secure_session  && !(w)->login.allow_access ?\
 				(w)->login.greeting : (w)->login.unsecure_greet)
-# define GREET_X(w)	((int)((w->core.width - \
+#define GREET_X(w)	((int)((w->core.width - \
 			     	STRING_WIDTH (greet, GREETING(w))) / 2))
-# define GREET_Y(w)	(GREETING(w)[0] ? 2 * GREET_Y_INC (w) : 0)
-#ifndef XPM
-# define GREET_W(w)	(max (STRING_WIDTH (greet, w->login.greeting), \
-			      STRING_WIDTH (greet, w->login.unsecure_greet)))
-#else
-# define GREET_W(w)	(max (STRING_WIDTH (greet, w->login.greeting), \
+#define GREET_Y(w)	(GREETING(w)[0] ? 2 * GREET_Y_INC (w) : 0)
+#define GREET_W(w)	(max (STRING_WIDTH (greet, w->login.greeting), \
 			      STRING_WIDTH (greet, w->login.unsecure_greet)) \
-			 + w->login.logoWidth + (2*w->login.logoPadding))
-#endif /* XPM */
-# define PROMPT_X(w)	(2 * PROMPT_X_INC(w))
-# define PROMPT_Y(w,n)	((GREET_Y(w) + GREET_Y_INC(w) +\
+			 + LOGO_W(w))
+
+#define PROMPT_X(w)	(2 * PROMPT_X_INC(w))
+#define PROMPT_Y(w,n)	((GREET_Y(w) + GREET_Y_INC(w) +\
 			  F_ASCENT(greet) + Y_INC(w)) + \
 			 (n * PROMPT_SPACE_Y(w)))
-# define PROMPT_W(w)	(w->core.width - (2 * TEXT_X_INC(w)))
-# define PROMPT_H(w)	(3 * Y_INC(w) / 2)
-# define VALUE_X(w,n)	(PROMPT_X(w) + CUR_PROMPT_W(w,n))
-# define PROMPT_SPACE_Y(w)	(10 * Y_INC(w) / 5)
+#define PROMPT_W(w)	(w->core.width - (2 * TEXT_X_INC(w)))
+#define PROMPT_H(w)	(5 * Y_INC(w) / 4)
+#define VALUE_X(w,n)	(PROMPT_X(w) + CUR_PROMPT_W(w,n))
+#define CURSOR_W	5
+#define MAX_VALUE_W(w,n) (PROMPT_W(w) - VALUE_X (w,n) - CURSOR_W - 1 - \
+			  (w->login.inframeswidth * 2) - LOGO_W(w))
+#define PROMPT_SPACE_Y(w)	(10 * Y_INC(w) / 5)
 
-# define ERROR_X(w,m)	((int)(w->core.width - STRING_WIDTH (fail, m)) / 2)
-# define FAIL_X(w)	ERROR_X(w, w->login.fail)
-# define FAIL_Y(w)	(PROMPT_Y(w,1) + 2 * FAIL_Y_INC (w) + F_ASCENT(fail))
+#define ERROR_X(w,m)	((int)(w->core.width - LOGO_W(w) - STRING_WIDTH (fail, m)) / 2)
+#define FAIL_X(w)	ERROR_X(w, w->login.fail)
+#define FAIL_Y(w)	(PROMPT_Y(w,1) + 2 * FAIL_Y_INC (w) + F_ASCENT(fail))
 
-#ifndef XPM
-# define ERROR_W(w,m)	STRING_WIDTH (fail, m)
-#else
-# define ERROR_W(w,m)	(STRING_WIDTH (fail, m) \
-			 + w->login.logoWidth + (2*w->login.logoPadding))
-#endif /* XPM */
+#define ERROR_W(w,m)	(STRING_WIDTH (fail, m) + LOGO_W(w))
 
-# define FAIL_W(w)	max(ERROR_W(w, w->login.failMsg), \
+#define FAIL_W(w)	max(ERROR_W(w, w->login.failMsg), \
 			    ERROR_W(w, w->login.passwdChangeMsg))
 
-# define PAD_X(w)	(2 * (PROMPT_X(w) + max (GREET_X_INC(w), FAIL_X_INC(w))))
+#define PAD_X(w)	(2 * (PROMPT_X(w) + max (GREET_X_INC(w), FAIL_X_INC(w))))
 
-# define PAD_Y(w)	(max (max (Y_INC(w), GREET_Y_INC(w)),\
+#define PAD_Y(w)	(max (max (Y_INC(w), GREET_Y_INC(w)),\
 			     FAIL_Y_INC(w)))
-	
+
 #ifndef max
 static inline int max (int a, int b) { return a > b ? a : b; }
 #endif
@@ -357,23 +355,40 @@ static inline int max (int a, int b) { return a > b ? a : b; }
 static void
 realizeValue (LoginWidget w, int cursor, int promptNum, GC gc)
 {
-    loginPromptState state = w->login.prompts[promptNum].state;
+    loginPromptState state = PROMPT_STATE(w, promptNum);
     char *text = VALUE_TEXT(w, promptNum);
     int	x, y, height, width, curoff;
 
     XDM_ASSERT(promptNum >= 0 && promptNum <= LAST_PROMPT);
 
+    /* replace all password characters with asterisks */
+    if ((state == LOGIN_PROMPT_ECHO_OFF) && (w->login.echo_passwd == True))
+    {
+	Cardinal length = strlen(text);
+	Cardinal i = 0;
+
+	text = XtMalloc(length + 1);
+
+	if (text == NULL)
+	{
+	    LogOutOfMem("realizeValue");
+	    return;
+	}
+
+	while (i < length)
+	{
+	    text[i++] = w->login.echo_passwd_char[0];
+	}
+
+	text[i] = 0;
+    }
+
     x = VALUE_X (w,promptNum);
     y = PROMPT_Y (w,promptNum);
 
-    height = PROMPT_H(w);
-    width = PROMPT_W(w) - x - 3;
+    height = PROMPT_H(w) - (w->login.inframeswidth * 2);
+    width = MAX_VALUE_W(w,promptNum);
 
-    height -= (w->login.inframeswidth * 2);
-    width -= (w->login.inframeswidth * 2);
-#ifdef XPM
-    width -= (w->login.logoWidth + 2*(w->login.logoPadding));
-#endif
     if (cursor > VALUE_SHOW_START(w, promptNum))
 	curoff = TEXT_WIDTH (text, text, cursor);
     else
@@ -386,26 +401,25 @@ realizeValue (LoginWidget w, int cursor, int promptNum, GC gc)
 			    x + curoff, y - TEXT_Y_INC(w),
 			    width - curoff, height);
 	}
-    } else if ((state == LOGIN_PROMPT_ECHO_ON) || (state == LOGIN_TEXT_INFO)) {
-	int textwidth;
+    } else if ((state == LOGIN_PROMPT_ECHO_ON) || (state == LOGIN_TEXT_INFO) ||
+	       ((state == LOGIN_PROMPT_ECHO_OFF) && (w->login.echo_passwd == True)))
+    {
 	int offset = max(cursor, VALUE_SHOW_START(w, promptNum));
 	int textlen = strlen (text + offset);
-	
-	textwidth = TEXT_WIDTH (text, text + offset, textlen);
 
-	if (textwidth > (width - curoff)) {
+	if (TEXT_WIDTH (text, text + offset, textlen) > (width - curoff)) {
 	    /* Recalculate amount of text that can fit in field */
 	    offset = VALUE_SHOW_START(w, promptNum);
 	    textlen = strlen (text + offset);
 
-	    while ((textlen > 0) && (textwidth > width))
+	    while ((textlen > 0) &&
+		   (TEXT_WIDTH (text, text + offset, textlen) > width))
 	    {
 		if (offset < PROMPT_CURSOR(w, promptNum)) {
 		    offset++;
 		}
 		textlen--;
-		textwidth = TEXT_WIDTH (text, text + offset, textlen);
-	    } 
+	    }
 
 	    VALUE_SHOW_START(w, promptNum) = offset;
 	    VALUE_SHOW_END(w, promptNum) = offset + textlen;
@@ -419,6 +433,11 @@ realizeValue (LoginWidget w, int cursor, int promptNum, GC gc)
 	    DRAW_STRING(text, x + curoff, y, text + offset, textlen);
 	}
     }
+    /* free memory */
+    if ((state == LOGIN_PROMPT_ECHO_OFF) && (w->login.echo_passwd == True))
+    {
+	XtFree(text);
+    }
 }
 
 static void
@@ -444,17 +463,10 @@ realizeCursor (LoginWidget w, GC gc)
     int	x, y;
     int height, width;
 
-#ifdef FORCE_CURSOR_FLASH    
-    static int lastx, lasty;
-    static struct timeval  lastFlash;
-    struct timeval  now, timeout;
-    int sinceLastFlash;
-#endif
-
     if (w->login.state != PROMPTING) {
 	return;
     }
-    
+
     x = VALUE_X (w, w->login.activePrompt);
     y = PROMPT_Y (w, w->login.activePrompt);
     height = (F_ASCENT(text) + F_DESCENT(text));
@@ -474,73 +486,61 @@ realizeCursor (LoginWidget w, GC gc)
 	}
 	break;
     case LOGIN_PROMPT_ECHO_OFF:
-	/* Nothing special needed */
+	if (w->login.echo_passwd == True) {
+	    if (w->login.echo_passwd_char[0] != 0) {
+		int len = PROMPT_CURSOR(w, w->login.activePrompt) -
+		    VALUE_SHOW_START(w, w->login.activePrompt);
+
+		x += len*TEXT_WIDTH(text, w->login.echo_passwd_char, 1);
+	    } else {
+		/* Move cursor one pixel per character to give some feedback
+		   without giving away the password length */
+		if (PROMPT_CURSOR(w, w->login.activePrompt) <
+		    MAX_VALUE_W(w, w->login.activePrompt))
+		    x += PROMPT_CURSOR(w, w->login.activePrompt);
+		else
+		    x += MAX_VALUE_W(w, w->login.activePrompt);
+	    }
+	}
 	break;
     }
-    
+
     XFillRectangle (XtDisplay (w), XtWindow (w), gc,
 		    x, y+1 - F_ASCENT(text), width, height-1);
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x-1 , y - F_ASCENT(text));
+		    x-1 , y - F_ASCENT(text));
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x+1 , y - F_ASCENT(text));
+		    x+1 , y - F_ASCENT(text));
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x-1 , y - F_ASCENT(text)+height);
+		    x-1 , y - F_ASCENT(text)+height);
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x+1 , y - F_ASCENT(text)+height);
+		    x+1 , y - F_ASCENT(text)+height);
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x-2 , y - F_ASCENT(text));
+		    x-2 , y - F_ASCENT(text));
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x+2 , y - F_ASCENT(text));
+		    x+2 , y - F_ASCENT(text));
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x-2 , y - F_ASCENT(text)+height);
+		    x-2 , y - F_ASCENT(text)+height);
     XDrawPoint     (XtDisplay (w), XtWindow (w), gc,
-    		    x+2 , y - F_ASCENT(text)+height);
+		    x+2 , y - F_ASCENT(text)+height);
 
-#ifdef FORCE_CURSOR_FLASH
-    /* Force cursor to flash briefly to give user feedback */
-#define FLASH_MILLIS    100000 /* 0.10 seconds */
-#define MILLIS_PER_SEC 1000000    
-    X_GETTIMEOFDAY (&now);
-
-    if ((lastx == x) && (lasty == y)) {
-	if (lastFlash.tv_sec == 0)
-	    sinceLastFlash = 0;
-	else if ((lastFlash.tv_sec + 1) == now.tv_sec)
-	    sinceLastFlash =
-		(now.tv_usec + MILLIS_PER_SEC) - lastFlash.tv_usec;
-	else if (lastFlash.tv_sec == now.tv_sec)
-	    sinceLastFlash = now.tv_usec - lastFlash.tv_usec;
-	else
-	    sinceLastFlash = (now.tv_sec - lastFlash.tv_sec) * MILLIS_PER_SEC;
-
-	if (sinceLastFlash < FLASH_MILLIS) {
-	    timeout.tv_sec = 0;
-	    timeout.tv_usec = FLASH_MILLIS - sinceLastFlash;
-	    select(0, NULL, NULL, NULL, &timeout);
-	}
-    } else {
-	lastx = x; lasty = y;
-    }
-    X_GETTIMEOFDAY (&lastFlash);
-#endif /* FORCE_CURSOR_FLASH */
-    XFlush (XtDisplay(w));    
+    XFlush (XtDisplay(w));
 }
 
 static void
 EraseFail (LoginWidget w)
 {
-#ifdef USE_XFT 
+#ifdef USE_XFT
     w->login.failUp = 0;
     RedrawFail(w);
-#else    
+#else
     XSetForeground (XtDisplay (w), w->login.failGC,
 			w->core.background_pixel);
     RedrawFail(w);
     w->login.failUp = 0;
     XSetForeground (XtDisplay (w), w->login.failGC,
 			TEXT_COLOR(fail));
-#endif    
+#endif
 }
 
 static void
@@ -592,9 +592,9 @@ RedrawFail (LoginWidget w)
     int y = FAIL_Y(w);
     int maxw = w->core.width - PAD_X(w);
 
-#ifndef USE_XFT    
+#ifndef USE_XFT
     if (w->login.failUp)
-#endif	
+#endif
     {
 	Debug("RedrawFail('%s', %d)\n", w->login.fail, w->login.failUp);
 	if (ERROR_W(w, w->login.fail) > maxw) {
@@ -602,10 +602,10 @@ RedrawFail (LoginWidget w)
 	    char *tempCopy = strdup(w->login.fail);
 	    if (tempCopy != NULL) {
 		char *start, *next;
-		char lastspace;
+		char lastspace = ' ';
 
 		y = PROMPT_Y(w,LAST_PROMPT) + (2 * PROMPT_Y_INC(w));
-		
+
 		for (start = next = tempCopy; start != NULL ; start = next) {
 		    /* search for longest string broken by whitespace that
 		       will fit on a single line */
@@ -625,14 +625,14 @@ RedrawFail (LoginWidget w)
 			    next = NULL;
 			}
 		    } while ((next != NULL) && ERROR_W(w, start) < maxw);
-		    
+
 		    x = ERROR_X(w, start);
 #ifdef USE_XFT
 		    if (w->login.failUp == 0) {
-			XClearArea(XtDisplay(w), XtWindow(w), x, y,
+			XClearArea(XtDisplay(w), XtWindow(w), x, y - F_ASCENT(fail),
 				   ERROR_W(w, start), FAIL_Y_INC(w), False);
 		    } else
-#endif		    
+#endif
 			DRAW_STRING (fail, x, y, start, strlen(start));
 
 		    if (next != NULL) {
@@ -650,10 +650,10 @@ RedrawFail (LoginWidget w)
 
 #ifdef USE_XFT
 	if (w->login.failUp == 0) {
-	    XClearArea(XtDisplay(w), XtWindow(w), x, y,
+	    XClearArea(XtDisplay(w), XtWindow(w), x, y - F_ASCENT(fail),
 		       ERROR_W(w, w->login.fail), FAIL_Y_INC(w), False);
 	} else
-#endif		    
+#endif
 	    DRAW_STRING (fail, x, y, w->login.fail, strlen (w->login.fail));
     }
 }
@@ -704,20 +704,19 @@ draw_it (LoginWidget w)
 
     EraseCursor (w);
 
-    if( (w->login.outframewidth) < 1 )
-      w->login.outframewidth = 1;
+    /* draw window borders */
     for(i=1;i<=(w->login.outframewidth);i++)
     {
       XDrawLine(XtDisplay (w), XtWindow (w), w->login.hiGC,
-    		i-1,i-1,w->core.width-i,i-1);
+		i-1,i-1,w->core.width-i,i-1);
       XDrawLine(XtDisplay (w), XtWindow (w), w->login.hiGC,
-    		i-1,i-1,i-1,w->core.height-i);
+		i-1,i-1,i-1,w->core.height-i);
       XDrawLine(XtDisplay (w), XtWindow (w), w->login.shdGC,
-    		w->core.width-i,i-1,w->core.width-i,w->core.height-i);
+		w->core.width-i,i-1,w->core.width-i,w->core.height-i);
       XDrawLine(XtDisplay (w), XtWindow (w), w->login.shdGC,
-    		i-1,w->core.height-i,w->core.width-i,w->core.height-i);
+		i-1,w->core.height-i,w->core.width-i,w->core.height-i);
     }
-    
+
     /* make separator line */
     gr_line_x = w->login.outframewidth;
     gr_line_y = GREET_Y(w) + GREET_Y_INC(w);
@@ -743,44 +742,47 @@ draw_it (LoginWidget w)
 	int in_frame_x = VALUE_X(w,p) - w->login.inframeswidth - 3;
 	int in_frame_y
 	    = PROMPT_Y(w,p) - w->login.inframeswidth - 1 - TEXT_Y_INC(w);
- 
-	int in_width = PROMPT_W(w) - VALUE_X(w,p);
-	int in_height = PROMPT_H(w) + w->login.inframeswidth + 2;
-	
-	GC topLeftGC, botRightGC;
 
-#ifdef XPM	
-	in_width -= (w->login.logoWidth + 2*(w->login.logoPadding));
-#endif /* XPM */
-	
+	int in_width = PROMPT_W(w) - VALUE_X(w,p) - LOGO_W(w);
+	int in_height = PROMPT_H(w) + w->login.inframeswidth + 2;
+
+	GC topLeftGC, botRightGC, inpGC;
+
 	if ((PROMPT_STATE(w, p) == LOGIN_PROMPT_ECHO_ON) ||
 	    (PROMPT_STATE(w, p) == LOGIN_PROMPT_ECHO_OFF)) {
 	    topLeftGC = w->login.shdGC;
 	    botRightGC = w->login.hiGC;
+	    inpGC = w->login.inpGC;
 	} else {
-	    topLeftGC = botRightGC = w->login.bgGC;
+	    topLeftGC = botRightGC = inpGC = w->login.bgGC;
 	}
 
+	/* draw borders of editboxes */
 	for (i=1; i<=(w->login.inframeswidth); i++)
 	{
 	    /* Make top/left sides */
 	    XDrawLine(XtDisplay (w), XtWindow (w), topLeftGC,
 		      in_frame_x + i-1,         in_frame_y + i-1,
-		      in_frame_x + in_width-i,  in_frame_y + i-1); 
+		      in_frame_x + in_width-i,  in_frame_y + i-1);
 
 	    XDrawLine(XtDisplay (w), XtWindow (w), topLeftGC,
 		      in_frame_x + i-1,         in_frame_y + i-1,
-		      in_frame_x + i-1,         in_frame_y + in_height-i); 
+		      in_frame_x + i-1,         in_frame_y + in_height-i);
 
 	    /* Make bottom/right sides */
 	    XDrawLine(XtDisplay (w), XtWindow (w), botRightGC,
 		      in_frame_x + in_width-i,  in_frame_y + i-1,
-		      in_frame_x + in_width-i,  in_frame_y + in_height-i); 
+		      in_frame_x + in_width-i,  in_frame_y + in_height-i);
 
 	    XDrawLine(XtDisplay (w), XtWindow (w), botRightGC,
 		      in_frame_x + i-1,         in_frame_y + in_height-i,
 		      in_frame_x + in_width-i,  in_frame_y + in_height-i);
 	}
+	XFillRectangle(XtDisplay (w), XtWindow (w), inpGC,
+	    in_frame_x + w->login.inframeswidth,
+	    in_frame_y + w->login.inframeswidth,
+	    in_width - 2*w->login.inframeswidth,
+	    in_height - 2*w->login.inframeswidth);
     }
 
     if (GREETING(w)[0]) {
@@ -788,7 +790,7 @@ draw_it (LoginWidget w)
 
 #ifdef XPM
 	gx -= ((w->login.logoWidth/2) + w->login.logoPadding);
-#endif	
+#endif
 	DRAW_STRING (greet, gx, GREET_Y(w), GREETING(w), strlen (GREETING(w)));
     }
     for (p = 0; p < NUM_PROMPTS ; p++) {
@@ -800,20 +802,8 @@ draw_it (LoginWidget w)
     }
     RedrawFail (w);
     XorCursor (w);
-    /*
-     * The GrabKeyboard here is needed only because of
-     * a bug in the R3 server -- the keyboard is grabbed on
-     * the root window, and the server won't dispatch events
-     * to the focus window unless the focus window is a ancestor
-     * of the grab window.  Bug in server already found and fixed,
-     * compatibility until at least R4.
-     */
-    if (XGrabKeyboard (XtDisplay (w), XtWindow (w), False, GrabModeAsync,
-		       GrabModeAsync, CurrentTime) != GrabSuccess)
-    {
-	XSetInputFocus (XtDisplay (w), XtWindow (w),
-			RevertToPointerRoot, CurrentTime);
-    }
+    XSetInputFocus (XtDisplay (w), XtWindow (w),
+		    RevertToPointerRoot, CurrentTime);
 }
 
 /* Returns 0 on success, -1 on failure */
@@ -829,12 +819,12 @@ SetPrompt (Widget ctx, int promptNum, const char *message,
 	"LOGIN_PROMPT_NOT_SHOWN", "LOGIN_PROMPT_ECHO_ON",
 	"LOGIN_PROMPT_ECHO_OFF", "LOGIN_TEXT_INFO" };
     loginPromptState priorState;
-    
+
     Debug("SetPrompt(%d, %s, %s(%d))\n", promptNum,
 	  message ? message : "<NULL>", stateNames[state], state);
 
     XDM_ASSERT(promptNum >= 0 && promptNum <= LAST_PROMPT);
-    
+
     if (PROMPT_TEXT(w, promptNum) != NULL) {
 	XtFree(PROMPT_TEXT(w, promptNum));
 	PROMPT_TEXT(w, promptNum) = NULL;
@@ -852,18 +842,18 @@ SetPrompt (Widget ctx, int promptNum, const char *message,
     }
 
     messageLen = strlen(message);
-    
+
     prompt = XtMalloc(messageLen + 3);
     if (prompt == NULL) {
 	LogOutOfMem ("SetPrompt");
 	return -1;
     }
-    
+
     strncpy(prompt, message, messageLen);
-    
+
     /* Make sure text prompts have at least two spaces at end */
     e = messageLen;
-	    
+
     if (!isspace(message[messageLen - 2])) {
 	prompt[e] = ' ';
 	e++;
@@ -879,11 +869,11 @@ SetPrompt (Widget ctx, int promptNum, const char *message,
     if (w->login.state == INITIALIZING) {
 	return 0;
     }
-    
+
     if ((priorState == LOGIN_TEXT_INFO) && (w->login.msgTimeout != 0)) {
 	time_t now = time(NULL);
 	int timeleft = w->login.msgTimeout - now;
-	
+
 	if (timeleft > 0) {
 	    sleep(timeleft);
 	}
@@ -900,7 +890,7 @@ SetPrompt (Widget ctx, int promptNum, const char *message,
 	w->login.activePrompt = promptNum;
 	w->login.state = PROMPTING;
     }
-    
+
     PROMPT_CURSOR(w, promptNum) = 0;
     XClearArea (XtDisplay(w), XtWindow(w), 0, 0, 0, 0, FALSE);
     draw_it(w);
@@ -914,7 +904,7 @@ GetPrompt(Widget ctx, int promptNum)
     LoginWidget	w = (LoginWidget) ctx;
 
     XDM_ASSERT(promptNum >= 0 && promptNum <= LAST_PROMPT);
-    
+
     return CUR_PROMPT_TEXT(w,promptNum);
 }
 
@@ -925,12 +915,12 @@ SetValue(Widget ctx, int promptNum, char *value)
     LoginWidget	w = (LoginWidget) ctx;
 
     XDM_ASSERT(promptNum >= 0 && promptNum <= LAST_PROMPT);
-    
+
     if ((promptNum < 0) || (promptNum > LAST_PROMPT))
 	return -1;
 
     XDM_ASSERT(VALUE_TEXT(w, promptNum) != NULL);
-    
+
     if (VALUE_TEXT(w, promptNum) == NULL)
 	return -1;
 
@@ -944,7 +934,7 @@ SetValue(Widget ctx, int promptNum, char *value)
     VALUE_SHOW_START(w, promptNum) = 0;
     VALUE_SHOW_END(w, promptNum) = 0;
     PROMPT_CURSOR(w, promptNum) = 0;
-    
+
     return 0;
 }
 
@@ -955,7 +945,7 @@ GetValue(Widget ctx, int promptNum)
     LoginWidget	w = (LoginWidget) ctx;
 
     XDM_ASSERT(promptNum >= 0 && promptNum <= LAST_PROMPT);
-    
+
     if ((promptNum < 0) || (promptNum > LAST_PROMPT))
 	return NULL;
 
@@ -971,7 +961,7 @@ realizeDeleteChar (LoginWidget ctx)
     if (ctx->login.state == PROMPTING) {
 	int promptNum = ctx->login.activePrompt;
 	int redrawFrom = PROMPT_CURSOR(ctx, promptNum);
-	
+
 	if (PROMPT_CURSOR(ctx,promptNum) <  (int)strlen(VALUE_TEXT(ctx,promptNum))) {
 	    if (redrawFrom < VALUE_SHOW_START(ctx, ctx->login.activePrompt)) {
 		redrawFrom = 0;
@@ -1005,7 +995,7 @@ DeleteBackwardChar (Widget ctxw, XEvent *event, String *params, Cardinal *num_pa
 	CUR_PROMPT_CURSOR(ctx) -= 1;
 	realizeDeleteChar(ctx);
     }
-    XorCursor (ctx);	
+    XorCursor (ctx);
 }
 
 /*ARGSUSED*/
@@ -1022,7 +1012,7 @@ DeleteForwardChar (Widget ctxw, XEvent *event, String *params, Cardinal *num_par
 
     XorCursor (ctx);
     realizeDeleteChar(ctx);
-    XorCursor (ctx);	
+    XorCursor (ctx);
 }
 
 /*ARGSUSED*/
@@ -1228,7 +1218,7 @@ TabField(Widget ctxw, XEvent *event, String *params, Cardinal *num_params)
 	if (nextPrompt > LAST_PROMPT) {
 	    nextPrompt = 0;
 	}
-    
+
 	if ((PROMPT_STATE(ctx, nextPrompt) == LOGIN_PROMPT_ECHO_ON) ||
 	    (PROMPT_STATE(ctx, nextPrompt) == LOGIN_PROMPT_ECHO_OFF)) {
 	    ctx->login.activePrompt = nextPrompt;
@@ -1446,7 +1436,7 @@ InsertChar (
 	    break;
     }
 
-    if (ctx->login.state == PROMPTING) {	
+    if (ctx->login.state == PROMPTING) {
 	if ((len + (int)strlen(VALUE_TEXT(ctx, promptNum)) >=
 	     (VALUE_TEXT_MAX(ctx,promptNum) - 1))) {
 	    len = VALUE_TEXT_MAX(ctx,promptNum) -
@@ -1481,7 +1471,7 @@ static XtConvertArgRec xftColorConvertArgs[] = {
      sizeof(Colormap)}
 };
 
-#define	donestr(type, value, tstr) \
+# define	donestr(type, value, tstr) \
 	{							\
 	    if (toVal->addr != NULL) {				\
 		if (toVal->size < sizeof(type)) {		\
@@ -1508,7 +1498,7 @@ XmuFreeXftColor (XtAppContext app, XrmValuePtr toVal, XtPointer closure,
     Screen	*screen;
     Colormap	colormap;
     XftColor	*color;
-    
+
     if (*num_args != 2)
     {
 	XtAppErrorMsg (app,
@@ -1527,7 +1517,7 @@ XmuFreeXftColor (XtAppContext app, XrmValuePtr toVal, XtPointer closure,
 				 XScreenNumberOfScreen (screen)),
 		  colormap, color);
 }
-    
+
 static Boolean
 XmuCvtStringToXftColor(Display *dpy,
 		       XrmValue *args, Cardinal *num_args,
@@ -1539,7 +1529,7 @@ XmuCvtStringToXftColor(Display *dpy,
     XftColor	    xftColor;
     Screen	    *screen;
     Colormap	    colormap;
-    
+
     if (*num_args != 2)
     {
 	XtAppErrorMsg (XtDisplayToApplicationContext (dpy),
@@ -1570,14 +1560,14 @@ XmuCvtStringToXftColor(Display *dpy,
     }
     else if (!XRenderParseColor (dpy, spec, &renderColor))
 	return False;
-    if (!XftColorAllocValue (dpy, 
+    if (!XftColorAllocValue (dpy,
 			     DefaultVisual (dpy,
 					    XScreenNumberOfScreen (screen)),
 			     colormap,
 			     &renderColor,
 			     &xftColor))
 	return False;
-    
+
     donestr (XftColor, xftColor, XtRXftColor);
 }
 
@@ -1587,7 +1577,7 @@ XmuFreeXftFont (XtAppContext app, XrmValuePtr toVal, XtPointer closure,
 {
     Screen  *screen;
     XftFont *font;
-    
+
     if (*num_args != 1)
     {
 	XtAppErrorMsg (app,
@@ -1613,7 +1603,7 @@ XmuCvtStringToXftFont(Display *dpy,
     char    *name;
     XftFont *font;
     Screen  *screen;
-    
+
     if (*num_args != 1)
     {
 	XtAppErrorMsg (XtDisplayToApplicationContext (dpy),
@@ -1626,7 +1616,7 @@ XmuCvtStringToXftFont(Display *dpy,
 
     screen = *((Screen **) args[0].addr);
     name = (char *) fromVal->addr;
-    
+
     font = XftFontOpenName (dpy,
 			    XScreenNumberOfScreen (screen),
 			    name);
@@ -1656,12 +1646,12 @@ XmuXftTextWidth(Display *dpy, XftFont *font, FcChar8 *string, int len)
 
 #endif /* USE_XFT */
 
-static void 
+static void
 ClassInitialize(void)
 {
 #ifdef USE_XFT
-    XtSetTypeConverter (XtRString, XtRXftColor, 
-			XmuCvtStringToXftColor, 
+    XtSetTypeConverter (XtRString, XtRXftColor,
+			XmuCvtStringToXftColor,
 			xftColorConvertArgs, XtNumber(xftColorConvertArgs),
 			XtCacheByDisplay, XmuFreeXftColor);
     XtSetTypeConverter (XtRString, XtRXftFont,
@@ -1689,7 +1679,7 @@ static void Initialize (
     int                 s_num;
 #endif
     int 	rv = 0;
-    
+
     myXGCV.foreground = w->login.hipixel;
     myXGCV.background = w->core.background_pixel;
     valuemask = GCForeground | GCBackground;
@@ -1700,6 +1690,11 @@ static void Initialize (
     valuemask = GCForeground | GCBackground;
     w->login.shdGC = XtGetGC(gnew, valuemask, &myXGCV);
 
+    myXGCV.foreground = w->login.inppixel;
+    myXGCV.background = w->core.background_pixel;
+    valuemask = GCForeground | GCBackground;
+    w->login.inpGC = XtGetGC(gnew, valuemask, &myXGCV);
+
     myXGCV.foreground = TEXT_COLOR(text);
     myXGCV.background = w->core.background_pixel;
     valuemask = GCForeground | GCBackground;
@@ -1708,12 +1703,12 @@ static void Initialize (
 	myXGCV.font = w->login.textFont->fid;
 	valuemask |= GCFont;
     }
-#endif    
+#endif
     w->login.textGC = XtGetGC(gnew, valuemask, &myXGCV);
-    myXGCV.foreground = w->core.background_pixel;
+    myXGCV.foreground = w->login.inppixel;
     w->login.bgGC = XtGetGC(gnew, valuemask, &myXGCV);
 
-    myXGCV.foreground = TEXT_COLOR(text) ^ w->core.background_pixel;
+    myXGCV.foreground = TEXT_COLOR(text) ^ w->login.inppixel;
     myXGCV.function = GXxor;
     xvaluemask = valuemask | GCFunction;
     w->login.xorGC = XtGetGC (gnew, xvaluemask, &myXGCV);
@@ -1740,7 +1735,7 @@ static void Initialize (
 
     xvaluemask = valuemask;
     if (w->login.greetFont == NULL)
-    	w->login.greetFont = w->login.textFont;
+	w->login.greetFont = w->login.textFont;
     else
 	xvaluemask |= GCFont;
 
@@ -1785,7 +1780,7 @@ static void Initialize (
 	     &(w->login.logoPixmap),			/* pixmap */
 	     &(w->login.logoMask),			/* pixmap mask */
 	     &myAttributes);				/* XPM attributes */
-       
+
 	if ( rv < 0 )
 	{
 	    LogError("Cannot load xpm file %s: %s.\n", w->login.logoFileName,
@@ -1827,7 +1822,7 @@ SkipXpmLoad:
     VALUE_TEXT(w, LOGIN_PROMPT_USERNAME) 	= w->login.data.name;
     VALUE_TEXT_MAX(w, LOGIN_PROMPT_USERNAME)	= sizeof(w->login.data.name);
     VALUE_SHOW_START(w, LOGIN_PROMPT_USERNAME)	= 0;
-    
+
     PROMPT_TEXT(w, LOGIN_PROMPT_PASSWORD) 	= NULL;
     DEF_PROMPT_TEXT(w, LOGIN_PROMPT_PASSWORD) 	= w->login.passwdPrompt;
     VALUE_TEXT(w, LOGIN_PROMPT_PASSWORD) 	= w->login.data.passwd;
@@ -1836,7 +1831,7 @@ SkipXpmLoad:
 
     SetPrompt(gnew, LOGIN_PROMPT_PASSWORD, NULL, LOGIN_PROMPT_ECHO_OFF, False);
     SetPrompt(gnew, LOGIN_PROMPT_USERNAME, NULL, LOGIN_PROMPT_ECHO_ON, False);
-    
+
     if (w->core.width == 0)
 	w->core.width = max (GREET_W(w), FAIL_W(w)) + PAD_X(w);
     if (w->core.height == 0) {
@@ -1850,7 +1845,7 @@ SkipXpmLoad:
 
         w->core.height = max(fy + pady,
             (w->login.logoHeight + (2*w->login.logoPadding)) + pady);
-        
+
 #endif /* XPM */
     }
 #ifdef USE_XINERAMA
@@ -1863,7 +1858,7 @@ SkipXpmLoad:
 	    x = screens[0].x_org + (int)(screens[0].width - w->core.width) / 2;
 	if ((y = w->core.y) == -1)
 	    y = screens[0].y_org + (int)(screens[0].height - w->core.height) / 3;
-	
+
 	XFree(screens);
     }
     else
@@ -1881,7 +1876,7 @@ SkipXpmLoad:
     w->login.state = PROMPTING;
 }
 
- 
+
 static void Realize (
      Widget gw,
      XtValueMask *valueMask,
@@ -1905,20 +1900,19 @@ static void Realize (
     XDefineCursor(XtDisplay(gw), DefaultRootWindow(XtDisplay(gw)), cursor);
 
 #ifdef XPM
-    /* 
+    /*
      * Check if Pixmap was valid
      */
     if (True == w->login.logoValid)
     {
         /*
-         * Create pixmap window 
+         * Create pixmap window
          */
         {
-            unsigned long valueMask = CWBackPixel | CWBackPixmap;
-            XSetWindowAttributes windowAttributes = { 0 };
-
-            windowAttributes.background_pixel = w->core.background_pixel;
-            windowAttributes.background_pixmap = None;
+            XSetWindowAttributes windowAttributes = {
+                .background_pixel = w->core.background_pixel,
+                .background_pixmap = None
+            };
 
             w->login.logoWindow = XCreateWindow(XtDisplay(w),
                 XtWindow(w),
@@ -1927,7 +1921,7 @@ static void Realize (
                 (w->core.height - w->login.logoHeight) /2,
                 w->login.logoWidth, w->login.logoHeight, 0,
                 CopyFromParent, InputOutput, CopyFromParent,
-                valueMask, &windowAttributes);
+                CWBackPixel | CWBackPixmap, &windowAttributes);
         }
 
         /*
@@ -1968,8 +1962,8 @@ static void Destroy (Widget gw)
 	XftDrawDestroy(w->login.draw);
 	w->login.draw = NULL;
     }
-#endif    
-    
+#endif
+
     XtReleaseGC(gw, w->login.textGC);
     XtReleaseGC(gw, w->login.bgGC);
     XtReleaseGC(gw, w->login.xorGC);
@@ -1980,6 +1974,7 @@ static void Destroy (Widget gw)
 #endif
     XtReleaseGC(gw, w->login.hiGC);
     XtReleaseGC(gw, w->login.shdGC);
+    XtReleaseGC(gw, w->login.inpGC);
 
 #ifdef XPM
     if (True == w->login.logoValid)
@@ -2011,7 +2006,7 @@ static Boolean SetValues (
     Cardinal *num_args)
 {
     LoginWidget currentL, newL;
-    
+
     currentL = (LoginWidget) current;
     newL = (LoginWidget) new;
     if (GREETING (currentL) != GREETING (newL))
